import logging

from cvxopt import matrix, mul, spmatrix, div

from .base import ModelBase
from ..consts import Gy  # NOQA
from ..utils.math import zeros

logger = logging.getLogger(__name__)


class Fault(ModelBase):
    """3-phase to ground fault class"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._group = 'Event'
        self._name = 'Fault'
        self._data.update({
            'bus': None,
            'fn': 60.0,
            'tf': None,
            'tc': None,
            'rf': 0,
            'xf': 1e-6,
        })
        self._params.extend([
            'bus',
            'fn',
            'tf',
            'tc',
            'rf',
            'xf',
        ])
        self._descr.update({
            'bus': 'bus number of fault',
            'fn': 'rated frequency',
            'tf': 'fault occurrence time',
            'tc': 'fault clearing time',
            'rf': 'fault resistance',
            'xf': 'fault impedance',
        })
        self._ac = {'bus': ['a', 'v']}
        self._z.extend(['rf', 'xf'])
        self._mandatory.extend(['bus', 'tf'])
        self._service = ['gf', 'bf', 'time', 'volt0', 'angle0']
        self.calls.update({'gcall': True, 'gycall': True})
        self._init()
        self.active = 0

    def setup(self):
        super().setup()
        self.xf += 1e-8
        Y = div(1, self.rf + 1j * self.xf)
        self.gf = Y.real()
        self.bf = Y.imag()
        self.u = zeros(self.n, 1)

    def get_times(self):
        if not self.n:
            return []
        t = matrix(list(self.tf) + list(self.tc))
        return list(t) + list(t - 1e-6)

    def is_time(self, t):
        return t in self.get_times()

    def apply(self, actual_time):
        """Check time and apply faults"""
        if self.time != actual_time:
            self.time = actual_time
        else:
            return

        for i in range(self.n):
            if self.tf[i] == self.time:
                logger.info(
                    ' <Fault> Applying fault on Bus <{}> at t={}.'.format(
                        self.bus[i], self.tf[i]))
                self.u[i] = 1
                self.active += 1
                self.angle0 = self.system.dae.y[self.system.Bus.a]
                self.volt0 = self.system.dae.y[self.system.Bus.n:]
                self.system.dae.factorize = True

            elif self.tc[i] == self.time:
                logger.info(
                    ' <Fault> Clearing fault on Bus <{}> at t={}.'.format(
                        self.bus[i], self.tc[i]))
                self.u[i] = 0
                self.active -= 1
                self.system.dae.y[self.system.Bus.n:] = self.volt0
                # self.system.dae.y[self.a] = self.anglepre
                self.system.dae.factorize = True

    def gcall(self, dae):
        if not self.active:
            return
        V2 = mul(self.u, dae.y[self.v]**2)
        p = mul(self.gf, V2)
        q = mul(self.bf, V2)
        dae.g += spmatrix(p, self.a, [0] * self.n, (dae.m, 1), 'd')
        dae.g -= spmatrix(q, self.v, [0] * self.n, (dae.m, 1), 'd')

    def gycall(self, dae):
        if not self.active:
            return
        V = mul(2, self.u, dae.y[self.v])
        dae.add_jac(Gy, mul(self.gf, V), self.a, self.v)
        dae.add_jac(Gy, -mul(self.bf, V), self.v, self.v)

    def insert(self, idx=None, name=None, **kwargs):
        if self.n:
            self._param_to_list()

        self.elem_add(idx, name, **kwargs)
        self.setup()



import logging

import numpy as np
from cvxopt import matrix, spmatrix, sparse, spdiag

from ..utils.math import ageb, aleb, aandb  # NOQA
from ..utils.math import index  # NOQA
from ..utils.math import zeros, ones

logger = logging.getLogger(__name__)


class DAE(object):
    """Class for numerical Differential Algebraic Equations (dae)"""

    def __init__(self, system):
        self.system = system
        self._data = dict(
            x=[],
            y=[],
            f=[],
            g=[],
            Fx=[],
            Fy=[],
            Gx=[],
            Gy=[],
            Fx0=[],
            Fy0=[],
            Gx0=[],
            Gy0=[],
            Ac=[],
            tn=[])

        self._scalars = dict(
            m=0,
            n=0,
            lamda=0,
            npf=0,
            kg=0,
            t=-1,
            factorize=True,
            rebuild=True,
            ac_reset=True)

        self._flags = dict(
            zxmax=[], zxmin=[], zymax=[], zymin=[], ux=[], uy=[])

        self.__dict__.update(self._data)
        self.__dict__.update(self._scalars)
        self.__dict__.update(self._flags)

        self._temp = {
            'Fx': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Fy': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Gx': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Gy': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Fx0': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Fy0': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Gx0': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Gy0': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
        }

        self._set = {
            'Fx': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Fy': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Gx': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Gy': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Fx0': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Fy0': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Gx0': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Gy0': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
        }
        self.Ac = []

    def init_xy(self):
        self.init_x()
        self.init_y()

    def init_x(self):
        self.x = zeros(self.n, 1)
        self.zxmax = ones(self.n, 1)
        self.zxmin = ones(self.n, 1)
        self.ux = ones(self.n, 1)

    def init_y(self):
        self.y = zeros(self.m, 1)
        self.zymax = ones(self.m, 1)
        self.zymin = ones(self.m, 1)
        self.uy = ones(self.m, 1)

    def init_fg(self, resetz=False):
        self.init_f(resetz)
        self.init_g(resetz)

    def init_f(self, resetz=True):
        self.zxmax = ones(self.n, 1)
        self.zxmin = ones(self.n, 1)
        self.f = zeros(self.n, 1)

    def init_g(self, resetz=True):
        self.zymax = ones(self.m, 1)
        self.zymin = ones(self.m, 1)
        self.g = zeros(self.m, 1)

    def setup_Gy(self):
        self.Gy = sparse(self.Gy0)
        self._temp.update({
            'Gy': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
        })

        self._set.update({
            'Gy': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
        })

    def setup_Fx(self):
        # self.Fx = sparse(self.Fx0)
        # self.Fy = sparse(self.Fy0)
        # self.Gx = sparse(self.Gx0)

        self._temp.update({
            'Fx': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Fy': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Gx': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
        })

        self._set.update({
            'Fx': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Fy': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Gx': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
        })

    def setup_FxGy(self):
        self.setup_Fx()
        self.setup_Gy()

    def setup(self):
        self.init_xy()
        self.init_fg()
        self.init_jac0()
        self.setup_Gy()
        self.setup_Fx()

    def init_Gy0(self):
        self.Gy0 = spmatrix([], [], [], (self.m, self.m), 'd')

        self._temp.update({
            'Gy0': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
        })

        self._set.update({
            'Gy0': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
        })

    def init_Fx0(self):
        self.Gx0 = spmatrix([], [], [], (self.m, self.n), 'd')
        self.Fy0 = spmatrix([], [], [], (self.n, self.m), 'd')
        self.Fx0 = spmatrix([], [], [], (self.n, self.n), 'd')

        self._temp.update({
            'Fx0': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Fy0': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Gx0': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
        })

        self._set.update({
            'Fx0': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Fy0': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
            'Gx0': {
                'I': matrix([]),
                'J': matrix([]),
                'V': matrix([])
            },
        })

    def init_jac0(self):
        self.init_Gy0()
        self.init_Fx0()

    def init1(self):
        self.resize()
        self.init_jac0()

    def resize(self):
        """Resize dae and and extend for init1 variables
        """
        yext = self.m - len(self.y)
        xext = self.n - len(self.x)

        if yext > 0:
            yzeros = zeros(yext, 1)
            yones = ones(yext, 1)
            self.y = matrix([self.y, yzeros], (self.m, 1), 'd')
            self.g = matrix([self.g, yzeros], (self.m, 1), 'd')
            self.uy = matrix([self.uy, yones], (self.m, 1), 'd')
            self.zymin = matrix([self.zymin, yones], (self.m, 1), 'd')
            self.zymax = matrix([self.zymax, yones], (self.m, 1), 'd')
        else:
            logger.debug('Skip extending DAE algeb arrays. Did you call `tds.init()` more than once or '
                         'have no dynamic components?')
        if xext > 0:
            xzeros = zeros(xext, 1)
            xones = ones(xext, 1)
            self.x = matrix([self.x, xzeros], (self.n, 1), 'd')
            self.f = matrix([self.f, xzeros], (self.n, 1), 'd')
            self.ux = matrix([self.ux, xones], (self.n, 1), 'd')
            self.zxmin = matrix([self.zxmin, xones], (self.n, 1), 'd')
            self.zxmax = matrix([self.zxmax, xones], (self.n, 1), 'd')
        else:
            logger.debug('Skip extending DAE state arrays. Did you call `tds.init()` more than once or '
                         'have no dynamic components?')

    def hard_limit(self, yidx, ymin, ymax, min_set=None, max_set=None):
        """Set hard limits for algebraic variables and reset the equation mismatches

        :param yidx: algebraic variable indices
        :param ymin: lower limit to check for
        :param ymax: upper limit to check for
        :param min_set: optional lower limit to set (``ymin`` as default)
        :param max_set: optional upper limit to set (``ymax`` as default)

        :type yidx: list, matrix
        :type ymin: matrix, int, float, list
        :type ymax: matrix, int, float, list
        :type min_set: matrix
        :type max_set: matrix

        :return: None

        """
        yidx = matrix(yidx)
        yval = self.y[yidx]
        ny = len(yidx)

        if isinstance(ymin, (int, float, list)):
            ymin = matrix(ymin, (ny, 1), 'd')
        if isinstance(ymax, (int, float, list)):
            ymax = matrix(ymax, (ny, 1), 'd')

        if not min_set:
            min_set = ymin
        elif isinstance(min_set, (int, float, list)):
            min_set = matrix(min_set, (ny, 1), 'd')

        if not max_set:
            max_set = ymax
        elif isinstance(max_set, (int, float, list)):
            max_set = matrix(max_set, (ny, 1), 'd')

        above = ageb(yval, ymax)
        below = aleb(yval, ymin)

        above_idx = index(above, 1.0)
        below_idx = index(below, 1.0)

        above_yidx = yidx[above_idx]
        below_yidx = yidx[below_idx]
        idx = list(above_idx) + list(below_idx)

        if len(above_yidx) > 0:
            self.y[above_yidx] = max_set[above_idx]
            self.zymax[above_yidx] = 0

        if len(below_yidx) > 0:
            self.y[below_yidx] = min_set[below_idx]
            self.zymin[below_yidx] = 0

        if len(idx):
            self.g[yidx[idx]] = 0
            self.ac_reset = True

    def hard_limit_remote(self, yidx, ridx, rtype='y', rmin=None, rmax=None, min_yset=0, max_yset=0):
        """Limit the output of yidx if the remote y is not within the limits

        TODO: This function needs to be refactored for clearliness.
        """
        ny = len(yidx)

        if ny != len(ridx):
            raise IndexError("len() of output vars and remote vars does not match")
        if rtype not in ('x', 'y'):
            raise TypeError('ridx has invalid type <{}>. Must be literal x or y'.format(rtype))

        if isinstance(min_yset, (int, float)):
            min_yset = matrix(min_yset, (ny, 1), 'd')
        if isinstance(max_yset, (int, float)):
            max_yset = matrix(max_yset, (ny, 1), 'd')

        above_idx, below_idx = list(), list()
        yidx = matrix(yidx)

        if rmax:
            # find the over-limit remote idx
            above = ageb(self.__dict__[rtype][ridx], rmax)
            above_idx = index(above, 1.0)
            # reset the y values based on the remote limit violations
            self.y[yidx[above_idx]] = max_yset[above_idx]
            self.zymax[yidx[above_idx]] = 0

        if rmin:
            below = aleb(self.__dict__[rtype][ridx], rmin)
            below_idx = index(below, 1.0)
            self.y[yidx[below_idx]] = min_yset[below_idx]
            self.zymin[yidx[below_idx]] = 0

        idx = above_idx + below_idx
        self.g[yidx[idx]] = 0

        if len(idx) > 0:
            self.factorize = True

    def anti_windup(self, xidx, xmin, xmax):
        """
        Anti-windup limiter for state variables.

        Resets the limited variables and differential equations.

        :param xidx: state variable indices
        :param xmin: lower limit
        :param xmax: upper limit

        :type xidx: matrix, list
        :type xmin: matrix, float, int, list
        :type xmax: matrix, float, int, list

        """
        xidx = matrix(xidx)
        xval = self.x[xidx]
        fval = self.f[xidx]

        if isinstance(xmin, (float, int, list)):
            xmin = matrix(xmin, xidx.size, 'd')

        if isinstance(xmax, (float, int, list)):
            xmax = matrix(xmax, xidx.size, 'd')

        x_above = ageb(xval, xmax)
        f_above = ageb(fval, 0.0)

        x_below = aleb(xval, xmin)
        f_below = aleb(fval, 0.0)

        above = aandb(x_above, f_above)
        above_idx = index(above, 1.0)
        if len(above_idx) > 0:
            above_xidx = xidx[above_idx]
            self.x[above_xidx] = xmax[above_idx]
            self.zxmax[above_xidx] = 0

        below = aandb(x_below, f_below)
        below_idx = index(below, 1.0)
        if len(below_idx) > 0:
            below_xidx = xidx[below_idx]
            self.x[below_xidx] = xmin[below_idx]
            self.zxmin[below_xidx] = 0

        idx = list(above_idx) + list(below_idx)
        if len(idx) > 0:
            self.f[xidx[idx]] = 0
            self.ac_reset = True

    def reset_Ac(self):
        """
        Reset ``dae.Ac`` sparse matrix for disabled equations
        due to hard_limit and anti_windup limiters.

        :return: None
        """
        if self.ac_reset is False:
            return

        mn = self.m + self.n
        x = index(aandb(self.zxmin, self.zxmax), 0.)
        y = [i + self.n for i in index(aandb(self.zymin, self.zymax), 0.)]
        xy = matrix(list(x) + y)
        n_xy = len(xy)

        if n_xy > 0:
            self.q[x] = 0

            if hasattr(self.Ac, 'ipset'):
                self.Ac[xy, :] = 0
                self.Ac[:, xy] = 0
                self.Ac.ipset(matrix(1, (n_xy, 1), 'd'), xy, xy)  # set the diagonals to 1
            else:
                eye = spdiag([1.0] * mn)
                # Modifying ``eye`` is more efficient than ``eye = eye - H``.
                # CVXOPT modifies eye in place because all the accessed elements exist.
                for idx in xy:
                    eye[idx, idx] = 0
                H = spmatrix(1.0, xy, xy, (mn, mn), 'd')
                self.Ac = eye * (self.Ac * eye) - H

        self.ac_reset = False
        self.factorize = True

    def get_size(self, m):
        """
        Return the 2-D size of a Jacobian matrix in tuple
        """
        nrow, ncol = 0, 0
        if m[0] == 'F':
            nrow = self.n
        elif m[0] == 'G':
            nrow = self.m

        if m[1] == 'x':
            ncol = self.n
        elif m[1] == 'y':
            ncol = self.m

        return nrow, ncol

    def add_jac(self, m, val, row, col):
        """Add tuples (val, row, col) to the Jacobian matrix ``m``

        Implemented in numpy.arrays for temporary storage.
        """
        if m not in ('Fx', 'Fy', 'Gx', 'Gy', 'Fx0', 'Fy0', 'Gx0', 'Gy0'):
            raise KeyError('Wrong Jacobian matrix name <{0}>'.format(m))

        if isinstance(val, (int, float)):
            val = val * ones(len(row), 1)

        self._temp[m]['I'] = matrix([self._temp[m]['I'], matrix(row)])
        self._temp[m]['J'] = matrix([self._temp[m]['J'], matrix(col)])
        self._temp[m]['V'] = matrix([self._temp[m]['V'], matrix(val)])

    def temp_to_spmatrix(self, ty):
        """
        Convert Jacobian tuples to matrices

        :param ty: name of the matrices to convert in ``('jac0','jac')``

        :return: None
        """
        if ty not in ('jac0', 'jac'):
            raise KeyError('Wrong type name <{}>. Must be literal jac0 or jac'.format(ty))

        jac0s = ['Fx0', 'Fy0', 'Gx0', 'Gy0']
        jacs = ['Fx', 'Fy', 'Gx', 'Gy']

        if ty == 'jac0':
            todo = jac0s
        elif ty == 'jac':
            todo = jacs

        for m in todo:
            self.__dict__[m] = spmatrix(self._temp[m]['V'],
                                        self._temp[m]['I'], self._temp[m]['J'],
                                        self.get_size(m), 'd')
            # add the jac0 elements
            if ty == 'jac':
                self.__dict__[m] += self.__dict__[m + '0']

        self.apply_set(ty)

    def set_jac(self, m, val, row, col):
        """
        Set the values at (row, col) to val in Jacobian m

        :param m: Jacobian name
        :param val: values to set
        :param row: row indices
        :param col: col indices
        :return: None
        """
        if m not in ('Fx', 'Fy', 'Gx', 'Gy', 'Fx0', 'Fy0', 'Gx0', 'Gy0'):
            raise KeyError('Wrong Jacobian matrix name <{0}>'.format(m))

        if isinstance(val, (int, float)) and isinstance(row, (np.ndarray, matrix, list)):
            val = val * ones(len(row), 1)

        self._set[m]['I'] = matrix([self._set[m]['I'], matrix(row)])
        self._set[m]['J'] = matrix([self._set[m]['J'], matrix(col)])
        self._set[m]['V'] = matrix([self._set[m]['V'], matrix(val)])

    def apply_set(self, ty):
        """
        Apply Jacobian set values to matrices

        :param ty: Jacobian type in ``('jac0', 'jac')``
        :return:
        """
        if ty not in ('jac0', 'jac'):
            raise KeyError("Wrong jacobian type {}. Must be literal jac0 or jac".format(ty))

        if ty == 'jac0':
            todo = ['Fx0', 'Fy0', 'Gx0', 'Gy0']
        else:
            todo = ['Fx', 'Fy', 'Gx', 'Gy']

        for m in todo:
            for idx in range(len(self._set[m]['I'])):
                i = self._set[m]['I'][idx]
                j = self._set[m]['J'][idx]
                v = self._set[m]['V'][idx]
                self.__dict__[m][i, j] = v

    def show(self, eq, value=None):
        """Show equation or variable array along with the names"""
        if eq in ['f', 'x']:
            key = 'unamex'
        elif eq in ['g', 'y']:
            key = 'unamey'

        if value:
            value = list(value)
        else:
            value = list(self.__dict__[eq])

        out = ''
        for name, val, idx in zip(self.system.varname.__dict__[key], value,
                                  range(len(value))):
            out += '{:20s} [{:>12.4f}] {:g}\n'.format(name, val, idx)
        return out

    def find_val(self, eq, val):
        """Return the name of the equation having the given value"""
        if eq not in ('f', 'g', 'q'):
            return
        elif eq in ('f', 'q'):
            key = 'unamex'
        elif eq == 'g':
            key = 'unamey'
        idx = 0
        for m, n in zip(self.system.varname.__dict__[key], self.__dict__[eq]):
            if n == val:
                return m, idx
            idx += 1
        return

    def reset_small(self, eq):
        """Reset numbers smaller than 1e-12 in f and g equations"""
        if eq not in ('f', 'g'):
            raise KeyError('Wrong equation type <{}>. Must be literal f or g'.format(eq))

        for idx, var in enumerate(self.__dict__[eq]):
            if abs(var) <= 1e-12:
                self.__dict__[eq][idx] = 0

    def reset_small_f(self):
        pass

    def reset_small_g(self):
        pass

    def check_diag(self, jac, name):
        """
        Check matrix ``jac`` for diagonal elements that equals 0
        """
        system = self.system
        pos = []
        names = []
        pairs = ''
        size = jac.size
        diag = jac[0:size[0] ** 2:size[0] + 1]

        for idx in range(size[0]):
            if abs(diag[idx]) <= 1e-8:
                pos.append(idx)

        for idx in pos:
            names.append(system.varname.__dict__[name][idx])

        if len(names) > 0:
            for i, j in zip(pos, names):
                pairs += '{0}: {1}\n'.format(i, j)
            logger.debug('Jacobian diagonal check:')
            logger.debug(pairs)

    # def add_jac(self, m, val, row, col):
    #     """Add values (val, row, col) to Jacobian m
    #
    #     This implementation construct the Jacobians incrementally.
    #       It is less efficient than storing (I,J,V) and create altogether.
    #     """
    #     if m not in ['Fx', 'Fy', 'Gx', 'Gy', 'Fx0', 'Fy0', 'Gx0', 'Gy0']:
    #         raise NameError('Wrong Jacobian matrix name <{0}>'.format(m))
    #
    #     size = self.__dict__[m].size
    #     self.__dict__[m] += spmatrix(val, row, col, size, 'd')
    #
    # def set_jac(self, m, val, row, col):
    #     """Add values (val, row, col) to Jacobian m
    #
    #     This implementation is very inefficient as it involves two
    #       __getitem__ and two __setitem__ ops.
    #     """
    #     if m not in ['Fx', 'Fy', 'Gx', 'Gy', 'Fx0', 'Fy0', 'Gx0', 'Gy0']:
    #         raise NameError('Wrong Jacobian matrix name <{0}>'.format(m))
    #
    #     old_val = []
    #     if type(row) is int:
    #         row = [row]
    #     if type(col) is int:
    #         col = [col]
    #     if type(row) is range:
    #         row = list(row)
    #     if type(col) is range:
    #         col = list(col)
    #     for i, j in zip(row, col):
    #         old_val.append(self.system.dae.__dict__[m][i, j])
    #     size = self.__dict__[m].size
    #     self.system.dae.__dict__[m] -= spmatrix(old_val, row, col, size, 'd')
    #     self.system.dae.__dict__[m] += spmatrix(val, row, col, size, 'd')
    #
    #
    # def hard_limit(self, yidx, ymin, ymax, min_set=None, max_set=None):
    #     """Limit algebraic variables and set the Jacobians.
    #
    #     This method is slower than the current version because,
    #       comparing with modifying matrix elements directly,
    #     matrix multiplication and re-assignment involves more ops.
    #     """
    #     yidx = matrix(yidx)
    #
    #     if not min_set:
    #         min_set = ymin
    #     if not max_set:
    #         max_set = ymax
    #
    #     if isinstance(min_set, (int, float)):
    #         min_set = matrix(min_set, (len(yidx), 1), 'd')
    #     if isinstance(max_set, (int, float)):
    #         max_set = matrix(max_set, (len(yidx), 1), 'd')
    #
    #     yvals = self.y[yidx]
    #
    #     above = ageb(yvals, ymax)
    #     below = aleb(yvals, ymin)
    #
    #     self.y[yidx] = mul(self.y[yidx], nota(above)) + mul(max_set, above)
    #     self.y[yidx] = mul(self.y[yidx], nota(below)) + mul(min_set, below)
    #
    #     self.zymax[yidx] = mul(self.zymax[yidx], nota(above))
    #     self.zymin[yidx] = mul(self.zymin[yidx], nota(below))
    #
    #     idx = aorb(above, below)
    #     self.g[yidx] = mul(self.g[yidx], nota(idx))
    #
    #     if sum(idx) > 0:
    #         self.ac_reset = True
    #
    #
    # def anti_windup(self, xidx, xmin, xmax):
    #     """State variable anti-windup limiter.
    #
    #     This version is slow for the same reason as hard_limit().
    #     """
    #     xidx = matrix(xidx)
    #     xval = self.x[xidx]
    #
    #     if type(xmin) in (float, int):
    #         xmin = matrix(xmin, xidx.size)
    #     if type(xmax) in (float, int):
    #         xmax = matrix(xmax, xidx.size)
    #
    #     x_above = ageb(xval, xmax)
    #     x_below = aleb(xval, xmin)
    #
    #     f_above = ageb(xval, 0.0)
    #     f_below = aleb(xval, 0.0)
    #
    #     above = aandb(x_above, f_above)
    #     below = aandb(x_below, f_below)
    #
    #     self.x[xidx] = mul(self.x[xidx], nota(above)) + mul(xmax, above)
    #     self.zxmax[xidx] = mul(self.zxmax[xidx], nota(above))
    #
    #     self.x[xidx] = mul(self.x[xidx], nota(below)) + mul(xmin, below)
    #     self.zxmin[xidx] = mul(self.zxmin[xidx], nota(below))
    #
    #     idx = aorb(above, below)
    #     self.f[xidx] = mul(self.f[xidx], nota(idx))
    #
    #     if sum(idx) > 0:
    #         self.factorize = True



import importlib
import logging
import os

from ..utils import elapsed

logger = logging.getLogger(__name__)
#
# Input formats is a dictionary of supported format names and the accepted
#   file extensions
#
# The first file will be parsed by read() function and the addfile will be
#   parsed by readadd()
#
# Typically, column based formats, such as IEEE CDF and PSS/E RAW,
#   are faster to parse
#
input_formats = {
    'dome': 'dm',
    'matpower': 'm',
    'psse': ['raw', 'dyr'],
    'card': ['andc'],
}

# Output formats is a dictionary of supported output formats and their
#   extensions
# The static data will be written by write() function and the addfile by
#   writeadd()
output_formats = ['']


def guess(system):
    """
    input format guess function. First guess by extension, then test by lines
    """
    files = system.files
    maybe = []
    if files.input_format:
        maybe.append(files.input_format)
    # first, guess by extension
    for key, val in input_formats.items():
        if isinstance(val, list):
            for item in val:
                if files.ext.strip('.').lower() == item:
                    maybe.append(key)
        else:
            if files.ext.strip('.').lower() == val:
                maybe.append(key)

    # second, guess by lines
    true_format = ''
    with open(files.case, 'r') as fid:
        for item in maybe:
            try:
                parser = importlib.import_module('.' + item, __name__)
                testlines = getattr(parser, 'testlines')
                if testlines(fid):
                    true_format = item
                    break
            except ImportError:
                logger.debug('Parser for {:s} format is not found. '
                             'Format guess will continue.'.format(item))

    if true_format:
        logger.debug('Input format guessed as {:s}.'.format(true_format))
    else:
        logger.error('Unable to determine case format.')

    files.input_format = true_format

    # guess addfile format
    if files.addfile:
        _, add_ext = os.path.splitext(files.addfile)
        for key, val in input_formats.items():
            if isinstance(val, list):
                if add_ext[1:] in val:
                    files.add_format = key
            else:
                if add_ext[1:] == val:
                    files.add_format = key

    return true_format


def parse(system):
    """
    Parse input file with the given format in system.files.input_format
    """

    t, _ = elapsed()

    input_format = system.files.input_format
    add_format = system.files.add_format
    # exit when no input format is given
    if not input_format:
        logger.error(
            'No input format found. Specify or guess a format before parsing.')
        return False

    # exit if the format parser could not be imported
    try:
        parser = importlib.import_module('.' + input_format, __name__)
        dmparser = importlib.import_module('.' + 'dome', __name__)
        if add_format:
            addparser = importlib.import_module('.' + add_format, __name__)
    except ImportError:
        logger.error(
            'Parser for {:s} format not found. Program will exit.'.format(
                input_format))
        return False

    # try parsing the base case file
    logger.info('Parsing input file <{:s}>'.format(system.files.fullname))

    if not parser.read(system.files.case, system):
        logger.error(
            'Error parsing case file {:s} with {:s} format parser.'.format(
                system.files.fullname, input_format))
        return False

    # Try parsing the addfile
    if system.files.addfile:
        if not system.files.add_format:
            logger.error('Unknown addfile format.')
            return
        logger.info('Parsing additional file {:s}.'.format(
            system.files.addfile))
        if not addparser.readadd(system.files.addfile, system):
            logger.error(
                'Error parsing addfile {:s} with {:s} format parser.'.format(
                    system.files.addfile, input_format))
            return False

    # Try parsing the dynfile with dm filter
    if system.files.dynfile:
        logger.info('Parsing input file {:s}.'.format(
            system.files.dynfile))
        if not dmparser.read(system.files.dynfile, system):
            logger.error(
                'Error parsing dynfile {:s} with dm format parser.'.format(
                    system.files.dynfile))
            return False

    _, s = elapsed(t)
    logger.debug('Case file {:s} parsed in {:s}.'.format(
        system.files.fullname, s))

    return True


def dump_raw(system):
    t, _ = elapsed()

    outfile = system.files.dump_raw
    dmparser = importlib.import_module('.' + 'dome', __name__)

    ret = dmparser.write(outfile, system)

    _, s = elapsed(t)
    if ret:
        logger.info('Raw file dump {:s} written in {:s}.'.format(
            system.files.dump_raw, s))
    else:
        logger.error('Dump raw file failed.')



"""Utility functions for loading andes stock test cases"""
import os


def stock_case_root():
    """Return the root path to the stock cases"""
    dir_name = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(dir_name, '..', '..', 'cases')


def get_stock_case(rpath):
    """Return the path to the stock cases"""
    case_root = stock_case_root()
    case_path = os.path.join(case_root, rpath)

    if not os.path.isfile(case_path):
        raise FileNotFoundError("Relative path {} is not valid for stock case".format(rpath))

    return case_path



#!/usr/bin/env python3

# ANDES, a power system simulation tool for research.
#
# Copyright 2015-2019 Hantao Cui
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Andes plotting tool
"""

import logging
import os
import re
import sys
from argparse import ArgumentParser
from distutils.spawn import find_executable

import numpy as np

logger = logging.getLogger(__name__)

try:
    from matplotlib import rc
    from matplotlib import pyplot as plt
except ImportError:
    logger.critical('Package <matplotlib> not found')
    sys.exit(1)

lfile = []
dfile = []


class TDSData(object):
    """
    A time-domain simulation data container for loading, extracing and
    plotting data
    """

    def __init__(self, file_name_full, path=None):
        # paths and file names
        self._path = path if path else os.getcwd()

        self.file_name, _ = os.path.splitext(file_name_full)

        self._dat_file = os.path.join(self._path, self.file_name + '.dat')
        self._lst_file = os.path.join(self._path, self.file_name + '.lst')

        # data members for raw data
        self._idx = []  # indices of variables
        self._uname = []  # unformatted variable names
        self._fname = []  # formatted variable names
        self._data = []  # data loaded from dat file

        # auxillary data members for fast query
        self.t = []
        self.nvars = 0  # total number of variables including `t`

        # TODO: consider moving the loading calls outside __init__
        self.load_lst()
        self.load_dat()

    def load_lst(self):
        """
        Load the lst file into internal data structures `_idx`, `_fname`, `_uname`, and counts the number of
        variables to `nvars`

        Returns
        -------
        None

        """
        with open(self._lst_file, 'r') as fd:
            lines = fd.readlines()

        idx, uname, fname = list(), list(), list()

        for line in lines:
            values = line.split(',')
            values = [x.strip() for x in values]

            # preserve the idx ordering here in case variables are not
            # ordered by idx
            idx.append(int(values[0]))  # convert to integer
            uname.append(values[1])
            fname.append(values[2])

        self._idx = idx
        self._fname = fname
        self._uname = uname
        self.nvars = len(uname)

    def find_var_idx(self, query, exclude=None, formatted=False):
        """
        Return variable names and indices matching `query`

        Parameters
        ----------
        query : str
            The string for querying variables
        exclude  : str, optional
            A string pattern to be excluded
        formatted : bool, optional
            True to return formatted names, False otherwise

        Returns
        -------
        (list, list)
            (List of found indices, list of found names)
        """

        # load the variable list to search in
        names = self._uname if formatted is False else self._fname

        found_idx, found_names = list(), list()

        for idx, name in zip(self._idx, names):
            if re.search(query, name):
                if exclude and re.search(exclude, name):
                    continue

                found_idx.append(idx)
                found_names.append(name)

        return found_idx, found_names

    def load_dat(self, delimiter=','):
        """
        Load the dat file into internal data structures `self._data`

        Parameters
        ----------
        delimiter : str, optional
            The delimiter for the case file. Default to comma.

        Returns
        -------
        None
        """
        try:
            data = np.loadtxt(self._dat_file, delimiter=',')
        except ValueError:
            data = np.loadtxt(self._dat_file)

        self._data = data

    def get_values(self, idx):
        """
        Return the variable values at the given indices

        Parameters
        ----------
        idx : list
            The indicex of the variables to retrieve. `idx=0` is for Time. Variable indices start at 1.

        Returns
        -------
        np.ndarray
            Variable data
        """
        return self._data[:, idx]

    def get_header(self, idx, formatted=False):
        """
        Return a list of the variable names at the given indices

        Parameters
        ----------
        idx : list or int
            The indices of the variables to retrieve
        formatted : bool
            True to retrieve latex-formatted names, False for unformatted names

        Returns
        -------
        list
            A list of variable names (headers)

        """

        if isinstance(idx, int):
            idx = [idx]
        header = self._uname if not formatted else self._fname
        return [header[x] for x in idx]

    def export_csv(self, path, idx=None, header=None, formatted=False,
                   sort_idx=True, fmt='%.18e'):
        """
        Export to a csv file

        Parameters
        ----------
        path : str
            path of the csv file to save
        idx : None or array-like, optional
            the indices of the variables to export. Export all by default
        header : None or array-like, optional
            customized header if not `None`. Use the names from the lst file
            by default
        formatted : bool, optional
            Use LaTeX-formatted header. Does not apply when using customized
            header
        sort_idx : bool, optional
            Sort by idx or not, # TODO: implement sort
        fmt : str
            cell formatter
        """
        if not idx:
            idx = self._idx
        if not header:
            header = self.get_header(idx, formatted=formatted)

        if len(idx) != len(header):
            raise ValueError("Idx length does not match header length")

        body = self.get_values(idx)

        with open(path, 'w') as fd:
            fd.write(','.join(header) + '\n')
            np.savetxt(fd, body, fmt=fmt, delimiter=',')

    def plot(self, yidx, xidx=(0,), y_calc=None, left=None, right=None, ymin=None, ymax=None, ytimes=None,
             xlabel=None, ylabel=None, legend=True, grid=False,
             latex=True, dpi=200, save=None, show=True, **kwargs):
        """
        Entery function for plot scripting. This function retrieves the x and y values based
        on the `xidx` and `yidx` inputs and then calls `plot_data()` to do the actual plotting.

        Note that `ytimes` and `y_calc` are applied sequentially if apply.

        Refer to `plot_data()` for the definition of arguments.

        Parameters
        ----------
        xidx : list or int
            The index for the x-axis variable

        yidx : list or int
            The indices for the y-axis variables

        Returns
        -------
        (fig, ax)
            Figure and axis handles
        """
        x_value = self.get_values(xidx)
        y_value = self.get_values(yidx)

        x_header = self.get_header(xidx, formatted=latex)
        y_header = self.get_header(yidx, formatted=latex)

        ytimes = float(ytimes) if ytimes is not None else ytimes

        if ytimes and (ytimes != 1):
            y_scale_func = scale_func(ytimes)
        else:
            y_scale_func = None

        # apply `ytimes` first
        if y_scale_func:
            y_value = y_scale_func(y_value)

        # `y_calc` is a callback function for manipulating data
        if y_calc is not None:
            y_value = y_calc(y_value)

        return self.plot_data(xdata=x_value, ydata=y_value, xheader=x_header, yheader=y_header,
                              left=left, right=right, ymin=ymin, ymax=ymax,
                              xlabel=xlabel, ylabel=ylabel, legend=legend, grid=grid,
                              latex=latex, dpi=dpi, save=save, show=show, **kwargs)

    def data_to_df(self):
        """Convert to pandas.DataFrame"""
        pass

    def guess_event_time(self):
        """Guess the event starting time from the input data by checking
        when the values start to change
        """
        pass

    def plot_data(self, xdata, ydata, xheader=None, yheader=None, xlabel=None, ylabel=None,
                  left=None, right=None, ymin=None, ymax=None, legend=True, grid=False, fig=None, ax=None,
                  latex=True, dpi=150, greyscale=False, save=None, show=True, **kwargs):
        """
        Plot lines for the supplied data and options. This functions takes `xdata` and `ydata` values. If
        you provide variable indices instead of values, use `plot()`.

        Parameters
        ----------
        xdata : array-like
            An array-like object containing the values for the x-axis variable

        ydata : array
            An array containing the values of each variables for the y-axis variable. The row
            of `ydata` must match the row of `xdata`. Each column correspondings to a variable.

        xheader : list
            A list containing the variable names for the x-axis variable

        yheader : list
            A list containing the variable names for the y-axis variable

        xlabel : str
            A label for the x axis

        ylabel : str
            A label for the y axis

        left : float
            The starting value of the x axis

        right : float
            The ending value of the x axis

        ymin : float
            The minimum value of the y axis
        ymax : float
            The maximum value of the y axis

        legend : bool
            True to show legend and False otherwise
        grid : bool
            True to show grid and False otherwise
        fig
            Matplotlib fig object to draw the axis on
        ax
            Matplotlib axis object to draw the lines on
        latex : bool
            True to enable latex and False to disable

        dpi : int
            Dots per inch for screen print or save
        greyscale : bool
            True to use greyscale, False otherwise
        save : bool
            True to save to png file
        show : bool
            True to show the image

        kwargs
            Optional kwargs

        Returns
        -------
        (fig, ax)
            The figure and axis handles
        """

        if not isinstance(ydata, np.ndarray):
            TypeError("ydata must be numpy array. Retrieve with get_values().")

        if ydata.ndim == 1:
            ydata = ydata.reshape((-1, 1))

        n_lines = ydata.shape[1]

        rc('font', family='Arial', size=12)

        using_latex = set_latex(latex)

        # set default x min based on simulation time
        if not left:
            left = xdata[0] - 1e-6
        if not right:
            right = xdata[-1] + 1e-6

        linestyles = ['-', '--', '-.', ':'] * len(ydata)

        if not (fig and ax):
            fig = plt.figure(dpi=dpi)
            ax = plt.gca()

        for i in range(n_lines):
            ax.plot(xdata, ydata[:, i],
                    ls=linestyles[i],
                    label=(yheader[i] if yheader else None),
                    linewidth=1,
                    )

        # for line, label in zip(line_objects, yheader):
        #     plt.legend(line, label)

        if xlabel:
            if using_latex:
                ax.set_xlabel(label_texify(xlabel))
        else:
            ax.set_xlabel(xheader[0])

        if ylabel:
            if using_latex:
                ax.set_ylabel(label_texify(ylabel))
            else:
                ax.set_ylabel(ylabel)

        ax.ticklabel_format(useOffset=False)

        ax.set_xlim(left=left, right=right)
        ax.set_ylim(ymin=ymin, ymax=ymax)

        if grid:
            ax.grid(b=True, linestyle='--')

        if legend:
            if yheader:
                ax.legend()

        plt.draw()

        if save:
            count = 1

            while True:
                outfile = self.file_name + '_' + str(count) + '.png'
                if not os.path.isfile(outfile):
                    break
                count += 1

            try:
                fig.savefig(outfile, dpi=dpi)
                logger.info('Figure saved to file {}'.format(outfile))
            except IOError:
                logger.error('* Error occurred. Try disabling LaTex with "-d".')
                return

        if show:
            plt.show()

        return fig, ax


def tdsplot_parse():
    """
    command line input parser for tdsplot

    Returns
    -------
    dict
        A dict of the command line arguments
    """
    parser = ArgumentParser(prog='tdsplot')
    parser.add_argument('datfile', nargs=1, default=[], help='dat file name.')
    parser.add_argument('x', nargs=1, type=int, help='x axis variable index')
    parser.add_argument('y', nargs='*', help='y axis variable index')
    parser.add_argument('--xmin', type=float, help='x axis minimum value', dest='left')
    parser.add_argument('--xmax', type=float, help='x axis maximum value', dest='right')
    parser.add_argument('--ymax', type=float, help='y axis maximum value')
    parser.add_argument('--ymin', type=float, help='y axis minimum value')

    parser.add_argument('--checkinit', action='store_true', help='check initialization value')

    parser.add_argument('-x', '--xlabel', type=str, help='manual x-axis text label')
    parser.add_argument('-y', '--ylabel', type=str, help='y-axis text label')

    parser.add_argument('-s', '--save', action='store_true', help='save to file')
    parser.add_argument('-g', '--grid', action='store_true', help='grid on')
    parser.add_argument('-d', '--no_latex', action='store_false', dest='latex', help='disable LaTex formatting')

    parser.add_argument('-n', '--no_show', action='store_false', dest='show', help='do not show the plot window')

    parser.add_argument('--ytimes', type=str, help='y times')
    parser.add_argument('--dpi', type=int, help='image resolution in dot per inch (DPI)')

    args = parser.parse_args()
    return vars(args)


def parse_y(y, upper, lower=0):
    """
    Parse command-line input for Y indices and return a list of indices

    Parameters
    ----------
    y : Union[List, Set, Tuple]
        Input for Y indices. Could be single item (with or without colon), or
         multiple items

    upper : int
        Upper limit. In the return list y, y[i] <= uppwer.

    lower : int
        Lower limit. In the return list y, y[i] >= lower.

    Returns
    -------

    """
    if len(y) == 1:
        if y[0].count(':') >= 3:
            logger.error('Index format not acceptable. Must not contain more than three colons.')
            return []

        elif y[0].count(':') == 0:
            if isint(y[0]):
                y[0] = int(y[0])
                return y
        elif y[0].count(':') == 1:
            if y[0].endswith(':'):
                y[0] += str(upper)
            if y[0].startswith(':'):
                y[0] = str(lower) + y[0]

        elif y[0].count(':') == 2:
            if y[0].endswith(':'):
                y[0] += str(1)
            if y[0].startswith(':'):
                y[0] = str(lower) + y[0]

            if y[0].count('::') == 1:
                y[0] = y[0].replace('::', ':{}:'.format(upper))
                print(y)

        y = y[0].split(':')

        for idx, item in enumerate(y[:]):
            try:
                y[idx] = int(item)
            except ValueError:
                logger.warning('y contains non-numerical values <{}>. Parsing could not proceed.'.format(item))
                return []

        y_from_range = list(range(*y))

        y_in_range = []

        for item in y_from_range:
            if lower <= item < upper:
                y_in_range.append(item)

        return y_in_range

    else:
        y_to_int = []
        for idx, val in enumerate(y):
            try:
                y_to_int.append(int(val))
            except ValueError:
                logger.warning('Y indices contains non-numerical values. Skipped <{}>.'.format(val))

        y_in_range = [item for item in y_to_int if lower <= item < upper]
        return list(y_in_range)


def add_plot(x, y, xl, yl, fig, ax, LATEX=False, linestyle=None, **kwargs):
    """Add plots to an existing plot"""
    if LATEX:
        # xl_data = xl[1]  # NOQA
        yl_data = yl[1]
    else:
        # xl_data = xl[0]  # NOQA
        yl_data = yl[0]

    for idx, y_val in enumerate(y):
        ax.plot(x, y_val, label=yl_data[idx], linestyle=linestyle)

    ax.legend(loc='upper right')
    ax.set_ylim(auto=True)


def eig_plot(name, args):
    fullpath = os.path.join(name, '.txt')
    raw_data = []
    started = 0
    fid = open(fullpath)
    for line in fid.readline():
        if '#1' in line:
            started = 1
        elif 'PARTICIPATION FACTORS' in line:
            started = -1

        if started == 1:
            raw_data.append(line)
        elif started == -1:
            break
    fid.close()

    for line in raw_data:
        # data = line.split()
        # TODO: complete this function
        pass


def tdsplot(datfile, y, x=(0,), **kwargs):
    """
    TDS plot main function based on the new TDSData class

    Parameters
    ----------
    datfile : str
        Path to the andes tds output data file (.dat or .lst file)
    x : list or int, optional
        The index for the x-axis variable. x=0 by default for time
    y : list or int
        The indices for the y-axis variable

    Returns
    -------
    TDSData object
    """

    # single data file
    if len(datfile) == 1:
        tds_data = TDSData(datfile[0])
        y_num = parse_y(y, lower=0, upper=tds_data.nvars)
        tds_data.plot(xidx=x, yidx=y_num, **kwargs)
        return tds_data
    else:
        raise NotImplementedError("Plotting multiple data files are not supported yet")


def tdsplot_main():
    """
    Entry function for tds plot. Parses command line arguments and calls `tdsplog`

    Returns
    -------
    None

    """
    args = tdsplot_parse()
    tdsplot(**args)


def check_init(yval, yl):
    """"Check initialization by comparing t=0 and t=end values"""
    suspect = []
    for var, label in zip(yval, yl):
        if abs(var[0] - var[-1]) >= 1e-6:
            suspect.append(label)
    if suspect:
        logger.error('Initialization failure:')
        logger.error(', '.join(suspect))
    else:
        logger.error('Initialization is correct.')


def isfloat(value):
    try:
        float(value)
        return True
    except ValueError:
        return False


def isint(value):
    try:
        int(value)
        return True
    except ValueError:
        return False


def scale_func(k):
    """
    Return a lambda function that scales its input by k

    Parameters
    ----------
    k : float
        The scaling factor of the returned lambda function
    Returns
    -------
    Lambda function

    """
    return lambda y_values_input: k * y_values_input


def label_texify(label):
    """
    Convert a label to latex format by appending surrounding $ and escaping spaces

    Parameters
    ----------
    label : str
        The label string to be converted to latex expression

    Returns
    -------
    str
        A string with $ surrounding
    """
    return '$' + label.replace(' ', r'\ ') + '$'


def set_latex(with_latex=True):
    """
    Enables latex for matplotlib based on the `with_latex` option and `dvipng` availability

    Parameters
    ----------
    with_latex : bool, optional
        True for latex on and False for off

    Returns
    -------
    bool
        True for latex on and False for off
    """

    has_dvipng = find_executable('dvipng')

    if has_dvipng and with_latex:
        rc('text', usetex=True)
        return True
    else:
        rc('text', usetex=False)
        return False


def main(cli=True, **args):
    logger.warning('andesplot is deprecated and will be remove in future versions. '
                   'Use "tdsplot" for TDS data or "eigplot" for EIG data.')
    tdsplot_main()


if __name__ == "__main__":
    main(cli=True)



import importlib
import logging
from math import ceil

import numpy.linalg
from cvxopt import matrix, spmatrix, mul, div
from cvxopt.lapack import gesv

from andes.config.eig import Eig
from andes.consts import pi
from andes.formats.txt import dump_data
from andes.utils import elapsed
from andes.utils.solver import Solver
from .base import RoutineBase

logger = logging.getLogger(__name__)
__cli__ = 'eig'


class EIG(RoutineBase):
    """
    Eigenvalue analysis routine
    """
    def __init__(self, system, rc=None):
        self.system = system
        self.solver = Solver(system.config.sparselib)
        self.config = Eig(rc=rc)

        # internal flags and storages
        self.As = None
        self.eigs = None
        self.mu = None
        self.part_fact = None

    def calc_state_matrix(self):
        """
        Return state matrix and store to ``self.As``

        Returns
        -------
        matrix
            state matrix
        """
        system = self.system

        Gyx = matrix(system.dae.Gx)
        self.solver.linsolve(system.dae.Gy, Gyx)

        self.As = matrix(system.dae.Fx - system.dae.Fy * Gyx)

        # ------------------------------------------------------
        # TODO: use scipy eigs
        # self.As = sparse(self.As)
        # I = np.array(self.As.I).reshape((-1,))
        # J = np.array(self.As.J).reshape((-1,))
        # V = np.array(self.As.V).reshape((-1,))
        # self.As = csr_matrix((V, (I, J)), shape=self.As.size)
        # ------------------------------------------------------
        return self.As

    def calc_eigvals(self):
        """
        Solve eigenvalues of the state matrix ``self.As``

        Returns
        -------
        None
        """
        self.eigs = numpy.linalg.eigvals(self.As)
        # TODO: use scipy.sparse.linalg.eigs(self.As)

        return self.eigs

    def calc_part_factor(self):
        """
        Compute participation factor of states in eigenvalues

        Returns
        -------

        """
        mu, N = numpy.linalg.eig(self.As)
        # TODO: use scipy.sparse.linalg.eigs(self.As)

        N = matrix(N)
        n = len(mu)
        idx = range(n)

        mu_complex = numpy.array([0] * n, dtype=complex)
        W = matrix(spmatrix(1.0, idx, idx, (n, n), N.typecode))
        gesv(N, W)

        partfact = mul(abs(W.T), abs(N))

        b = matrix(1.0, (1, n))
        WN = b * partfact
        partfact = partfact.T

        for item in idx:
            mu_real = float(mu[item].real)
            mu_imag = float(mu[item].imag)
            mu_complex[item] = complex(round(mu_real, 4), round(mu_imag, 4))
            partfact[item, :] /= WN[item]

        # participation factor
        self.mu = matrix(mu_complex)
        self.part_fact = matrix(partfact)

        return self.mu, self.part_fact

    def run(self, **kwargs):
        ret = False
        system = self.system

        if system.pflow.solved is False:
            logger.warning(
                'Power flow not solved. Eig analysis will not continue.')
            return ret
        elif system.dae.n == 0:
            logger.warning('No dynamic model. Eig analysis will not continue.')
            return ret

        t1, s = elapsed()
        logger.info('-> Eigenvalue Analysis:')

        system.dae.factorize = True
        system.call.int()

        self.calc_state_matrix()
        self.calc_part_factor()

        self.dump_results()
        self.plot_results()
        ret = True

        _, s = elapsed(t1)
        logger.info('Eigenvalue analysis finished in {:s}.'.format(s))

        return ret

    def plot_results(self):
        try:
            plt = importlib.import_module('matplotlib.pyplot')
        except ImportError:
            plt = None
            logger.warning('Install matplotlib to plot eigenvalue map.')
            return False

        mu_real = self.mu.real()
        mu_imag = self.mu.imag()
        p_mu_real, p_mu_imag = list(), list()
        z_mu_real, z_mu_imag = list(), list()
        n_mu_real, n_mu_imag = list(), list()

        for re, im in zip(mu_real, mu_imag):
            if re == 0:
                z_mu_real.append(re)
                z_mu_imag.append(im)
            elif re > 0:
                p_mu_real.append(re)
                p_mu_imag.append(im)
            elif re < 0:
                n_mu_real.append(re)
                n_mu_imag.append(im)

        if len(p_mu_real) > 0:
            logger.warning(
                'System is not stable due to {} positive eigenvalues.'.format(
                    len(p_mu_real)))
        else:
            logger.info(
                'System is small-signal stable in the initial neighbourhood.')

        if self.config.plot:
            fig, ax = plt.subplots()
            ax.scatter(n_mu_real, n_mu_imag, marker='x', s=26, color='green')
            ax.scatter(z_mu_real, z_mu_imag, marker='o', s=26, color='orange')
            ax.scatter(p_mu_real, p_mu_imag, marker='x', s=26, color='red')
            plt.show()

    def dump_results(self):
        """
        Save eigenvalue analysis reports

        Returns
        -------
        None
        """
        system = self.system
        mu = self.mu
        partfact = self.part_fact

        if system.files.no_output:
            return

        text = []
        header = []
        rowname = []
        data = []

        neig = len(mu)
        mu_real = mu.real()
        mu_imag = mu.imag()
        npositive = sum(1 for x in mu_real if x > 0)
        nzero = sum(1 for x in mu_real if x == 0)
        nnegative = sum(1 for x in mu_real if x < 0)

        numeral = []
        for idx, item in enumerate(range(neig)):
            if mu_real[idx] == 0:
                marker = '*'
            elif mu_real[idx] > 0:
                marker = '**'
            else:
                marker = ''
            numeral.append('#' + str(idx + 1) + marker)

        # compute frequency, undamped frequency and damping
        freq = [0] * neig
        ufreq = [0] * neig
        damping = [0] * neig
        for idx, item in enumerate(mu):
            if item.imag == 0:
                freq[idx] = 0
                ufreq[idx] = 0
                damping[idx] = 0
            else:
                freq[idx] = abs(item) / 2 / pi
                ufreq[idx] = abs(item.imag / 2 / pi)
                damping[idx] = -div(item.real, abs(item)) * 100

        # obtain most associated variables
        var_assoc = []
        for prow in range(neig):
            temp_row = partfact[prow, :]
            name_idx = list(temp_row).index(max(temp_row))
            var_assoc.append(system.varname.unamex[name_idx])

        pf = []
        for prow in range(neig):
            temp_row = []
            for pcol in range(neig):
                temp_row.append(round(partfact[prow, pcol], 5))
            pf.append(temp_row)

        text.append(system.report.info)
        header.append([''])
        rowname.append(['EIGENVALUE ANALYSIS REPORT'])
        data.append('')

        text.append('STATISTICS\n')
        header.append([''])
        rowname.append(['Positives', 'Zeros', 'Negatives'])
        data.append([npositive, nzero, nnegative])

        text.append('EIGENVALUE DATA\n')
        header.append([
            'Most Associated', 'Real', 'Imag', 'Damped Freq.', 'Frequency',
            'Damping [%]'
        ])
        rowname.append(numeral)
        data.append(
            [var_assoc,
             list(mu_real),
             list(mu_imag), ufreq, freq, damping])

        cpb = 7  # columns per block
        nblock = int(ceil(neig / cpb))

        if nblock <= 100:
            for idx in range(nblock):
                start = cpb * idx
                end = cpb * (idx + 1)
                text.append('PARTICIPATION FACTORS [{}/{}]\n'.format(
                    idx + 1, nblock))
                header.append(numeral[start:end])
                rowname.append(system.varname.unamex)
                data.append(pf[start:end])

        dump_data(text, header, rowname, data, system.files.eig)
        logger.info('report saved.')



import logging

from cvxopt import spmatrix, spdiag, matrix

from .base import ModelBase
from ..utils.math import zeros

logger = logging.getLogger(__name__)


class Zone(ModelBase):
    """Zone class"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._group = 'Topo_Zone'
        self._name = 'Zone'
        self._params.extend(['pdes', 'ptol', 'isw'])
        self._descr.update({
            'pdes':
            "Desired net interchange leaving this area",
            'ptol':
            'Interchange tolerance bandwidth',
            'isw':
            'Area slack bus idx',
        })
        self._units.update({
            'pdes': 'MW',
            'ptol': 'MW',
        })
        self._data.update({
            'pdes': 0.0,
            'ptol': 10.0,
            'isw': 0,
        })
        self._powers.extend(['pdes', 'ptol'])
        self.calls.update({'pflow': True})
        self._service.extend(['uname', 'fname'])
        self._init()

        # shape of `self.incidence` is (self.n, self.n) with
        #   the diagonals are the number of lines in the area, and
        #   the off-diagonals are the number of tie-lines between area  `row`
        #  and `col`

        self.incidence = None
        self.tielines = dict()  # [tielines, boarder bus] in each area
        self.buses = dict()  # buses in each area
        # self.area_flows = dict()  # tieline outgoings [P, Q] for each area
        self.area_P0 = []
        self.area_Q0 = []
        self.interchange = dict(
        )  # a sparse matrix with the interchange values
        self.inter_flows = dict()
        self.inter_varout = matrix([])

        self.inter_uname = list()
        self.inter_fname = list()
        self.n_combination = 0

    def setup(self):
        if not self.n:
            return
        # TODO: account for >1 area/region/zone
        super().setup()
        var = self._name.lower()
        for idx, uid in self.system.Bus.uid.items():
            code = self.system.Bus.__dict__[var][uid]
            if code and code not in self.idx:
                logger.warning('{} <{}> not defined.'.format(
                    self._name, code))
            if code not in self.buses.keys():
                self.buses[code] = list()
            self.buses[code].append(idx)

        x, a, b = list(), list(), list()
        for idx, uid in self.system.Line.uid.items():
            bus1 = self.system.Line.bus1[uid]
            bus2 = self.system.Line.bus2[uid]
            code1 = self.system.Bus.__dict__[var][self.system.Bus.uid[bus1]]
            code2 = self.system.Bus.__dict__[var][self.system.Bus.uid[bus2]]

            if code1 == 0 or code2 == 0:
                continue
            if code1 != code2:
                if code1 not in self.tielines.keys():
                    self.tielines[code1] = list()
                if code2 not in self.tielines.keys():
                    self.tielines[code2] = list()

                self.tielines[code1].append([idx, bus1])
                self.tielines[code2].append([idx, bus2])

            if code1 != code2:
                if code1 not in self.interchange:
                    self.interchange[code1] = dict()
                if code2 not in self.interchange:
                    self.interchange[code2] = dict()
                if code2 not in self.interchange[code1]:
                    self.interchange[code1][code2] = []
                if code1 not in self.interchange[code2]:
                    self.interchange[code2][code1] = []

                self.interchange[code1][code2].append([idx, bus1])
                self.interchange[code2][code1].append([idx, bus2])

            int_code1 = self.uid[code1]
            int_code2 = self.uid[code2]

            if int_code1 == int_code2:
                x.append(0.5)
            else:
                x.append(1)

            a.append(int_code1)
            b.append(int_code2)

        half_incidence = spmatrix(x, a, b, (self.n, self.n), 'd')
        self.incidence = half_incidence + half_incidence.T

        incidence = self.incidence - spdiag(self.incidence[0::self.n + 1])
        I, J, V = incidence.I, incidence.J, incidence.V

        self.n_combination = 0
        for i, j, v in zip(I, J, V):
            if not v:
                continue
            self.n_combination += 1
            self.uname.append('{}-{}'.format(self.name[i], self.name[j]))
            self.fname.append('{}-{}'.format(self.name[i], self.name[j]))

        self.area_P0 = zeros(self.n, 1)
        self.area_Q0 = zeros(self.n, 1)

    def seriesflow(self, dae):
        for code in self.idx:
            int_idx = self.uid[code]
            pairs = self.tielines.get(code, None)
            if not pairs:
                continue
            lines = [i for i, _ in pairs]
            buses = [j for _, j in pairs]

            P0, Q0 = self.system.Line.get_flow_by_idx(idx=lines, bus=buses)
            self.area_P0[int_idx] = sum(P0)
            self.area_Q0[int_idx] = sum(Q0)

        for fr_area, pairs in self.interchange.items():
            if fr_area not in self.inter_flows.keys():
                self.inter_flows[fr_area] = dict()

            for to_area, vals in pairs.items():
                lines = [i for i, _ in vals]
                buses = [j for _, j in vals]
                P, Q = self.system.Line.get_flow_by_idx(idx=lines, bus=buses)
                self.inter_flows[fr_area][to_area] = [sum(P), sum(Q)]

        self.interchange_varout()

    def _varname_inter(self):
        if not self.n:
            return
        mpql = self.system.dae.m + 2 * self.system.Bus.n + 8 * self.system.Line.n

        # P_ic
        xy_idx = range(mpql, mpql + self.n_combination)
        self.system.varname.append(
            listname='unamey',
            xy_idx=xy_idx,
            var_name='P_ic',
            element_name=self.uname)
        self.system.varname.append(
            listname='fnamey',
            xy_idx=xy_idx,
            var_name='P_{ic}',
            element_name=self.fname)

        # Q_ic
        xy_idx = range(mpql + self.n_combination,
                       mpql + 2 * self.n_combination)
        self.system.varname.append(
            listname='unamey',
            xy_idx=xy_idx,
            var_name='Q_ic',
            element_name=self.uname)
        self.system.varname.append(
            listname='fnamey',
            xy_idx=xy_idx,
            var_name='Q_{ic}',
            element_name=self.fname)

    def interchange_varout(self):
        if not self.n:
            return
        incidence = self.incidence - spdiag(self.incidence[0::self.n + 1])
        I, J, V = incidence.I, incidence.J, incidence.V
        P, Q = [], []
        for i, j, v in zip(I, J, V):
            if v == 0.:
                continue
            P.append(self.inter_flows[self.idx[i]][self.idx[j]][0])
            Q.append(self.inter_flows[self.idx[i]][self.idx[j]][1])

        self.P = matrix(P)
        self.Q = matrix(Q)
        self.inter_varout = matrix(P + Q)


class Area(Zone):
    """Area class"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._name = 'Area'
        self._group = 'Topo_Area'
        self._init()

    def setup(self):
        # TODO: account for >1 area/region/zone
        super().setup()


class Region(Zone):
    """Region class"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._name = 'Region'
        self._group = 'Topo_Region'
        self._params.extend(['Ptol', 'slack'])
        self._descr.update({
            'Ptol': 'Total transfer capacity',
            'slack': 'slack bus idx',
        })
        self._data.update({
            'Ptol': None,
            'slack': None,
        })
        self._powers.extend(['Ptol'])
        self._init()

    def setup(self):
        super().setup()
